import asyncio
import oneai
from typing import List, Optional, Union
from typing_extensions import Literal
from dataclasses import dataclass


@dataclass
class Chapter:
    subheading: str
    text: Union[str, List[oneai.Utterance]]
    summary: str


def span_text(
    input: Union[str, List[oneai.Utterance]], label: oneai.Label
) -> Union[str, List[oneai.Utterance]]:
    if isinstance(input, str):
        return label.span_text
    return [
        oneai.Utterance(
            speaker=input[span.section].speaker,
            utterance=input[span.section].utterance[span.start : span.end],
            timestamp=input[span.section].timestamp,
        )
        for span in label.output_spans
    ]


def generate_chapters(
    input: oneai.PipelineInput,
    amount: Literal["more", "less", "normal"] = None,
    preprocessing: Optional[oneai.Skill] = None,
) -> List[Chapter]:
    output = asyncio.run(
        oneai.Pipeline([oneai.skills.SplitByTopic(amount=amount)]).run_async(input)
        if not preprocessing
        else oneai.Pipeline(
            [preprocessing, oneai.skills.SplitByTopic(amount=amount)]
        ).run_async(input)
    )
    if preprocessing:
        output = output.__getattribute__(preprocessing.output_attr)
    spans = [span_text(output.text, seg) for seg in output.segments]
    summaries = oneai.Pipeline([oneai.skills.Summarize()]).run_batch(spans)
    return [
        Chapter(
            subheading=seg.data["subheading"],
            text=span,
            summary=summaries[span].summary.text,
        )
        for seg, span in zip(output.segments, spans)
    ]



from oneai.api.pipeline import post_pipeline
from oneai.api.clustering import (
    post_clustering,
    get_clustering,
    get_clustering_paginated,
)


def get_or_create_uuid() -> str:
    import os, uuid

    with open(f"{os.path.dirname(__file__)}/.uuid", "a+") as f:
        f.seek(0)
        id = f.read()
        if not id:
            id = uuid.uuid4().hex
            f.write(id)
        return id


uuid = get_or_create_uuid()



import asyncio
import concurrent.futures
import io
import os
import sys
from typing import Awaitable, Callable, Dict, Iterable, List

import oneai
from oneai.classes import BatchResponse, Output, PipelineInput, Skill, TextContent
from oneai.process_scheduler import *


class Pipeline:
    """
    Language AI pipelines allow invoking and chaining multiple Language Skills to process your input text with a single API call.

    ## Attributes

    `steps: list[Skill]`
        A list of Language Skills to process the input text. The order of the skills in the list determines their input.
    `api_key: str, optional`
        An API key to be used in this pipelines `run` calls. If not provided, the global `oneai.api_key` is used.
    `multilingual: bool, optional`
        Whether the pipeline should be allowed to process multilingual input.

    ## Methods

    `run(input, api_key=None) -> Output`
        Runs the pipeline on the input text.
    `run_async(input, api_key=None) -> Awaitable[Output]`
        Runs the pipeline on the input text asynchronously.
    `run_batch(batch, api_key=None) -> Dict[Input, Output]`
        Runs the pipeline on a batch of input texts.
    `run_batch_async(batch, api_key=None) -> Awaitable[Dict[Input, Output]]`
        Runs the pipeline on a batch of input texts asynchronously.

    ## Pipeline Ordering

    The order of the skills in the pipeline determines their input:
    * Skills that are placed after a generator Skill will receive its output as input.
    * If there's no preceding generator Skill, the original input text is used.

    ## Example

    >>> my_text = 'ENTER-YOUR-TEXT-HERE'
    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Topics(),
    ...     oneai.skills.Summarize(min_length=20),
    ...     oneai.skills.Entities()
    ... ])
    >>> output = pipeline.run(my_text)
    >>> output.topics
    [oneai.Label(type=topic, span=[0, 10], name=topic1), ...] # topics from my_text
    >>> output.summary.text
    '...' # summary of my_text
    >>> output.summary.entities
    [oneai.Label(type=entity, span=[0, 10], name=entity1), ...] # entities from the summary
    """

    def __init__(
        self, steps: List[Skill], api_key: str = None, multilingual: bool = False
    ) -> None:
        self.steps = tuple(steps)  # todo: validate (based on input_type)
        self.api_key = api_key
        self.multilingual = multilingual

    def run(
        self,
        input: PipelineInput[TextContent],
        api_key: str = None,
        multilingual: bool = False,
    ) -> Output[TextContent]:
        """
        Runs the pipeline on the input text.

        ## Parameters

        `input: PipelineInput`
            The input text to be processed.
        `api_key: str, optional`
            An API key to be used in this API call. If not provided, `self.api_key` is used.

        ## Returns

        An `Output` object containing the results of the Skills in the pipeline.

        ## Raises

        `InputError` if the input is is invalid or is of an incompatible type for the pipeline.
        `APIKeyError` if the API key is invalid, expired, or missing quota.
        `ServerError` if an internal server error occured.
        """
        return _async_run_nested(
            process_single_input(
                input,
                self.steps,
                api_key or self.api_key or oneai.api_key,
                multilingual or self.multilingual or oneai.multilingual,
            )
        )

    async def run_async(
        self,
        input: PipelineInput[TextContent],
        api_key: str = None,
        interval: int = 1,
        multilingual: bool = False,
    ) -> Awaitable[Output[TextContent]]:
        """
        Runs the pipeline on the input text asynchronously.

        ## Parameters

        `input: PipelineInput`
            The input text (or multiple input texts) to be processed.
        `api_key: str, optional`
            An API key to be used in this API call. If not provided, `self.api_key` is used.

        ## Returns

        An Awaitable with an `Output` object containing the results of the Skills in the pipeline.

        ## Raises

        `InputError` if the input is is invalid or is of an incompatible type for the pipeline.
        `APIKeyError` if the API key is invalid, expired, or missing quota.
        `ServerError` if an internal server error occured.
        """
        return await (
            process_file_async(
                input,
                self.steps,
                api_key or self.api_key or oneai.api_key,
                interval,
                multilingual or self.multilingual or oneai.multilingual,
            )
            if isinstance(input, io.IOBase)
            or (isinstance(input, Input) and isinstance(input.text, io.IOBase))
            else process_single_input(
                input,
                self.steps,
                api_key or self.api_key or oneai.api_key,
                multilingual or self.multilingual or oneai.multilingual,
            )
        )

    def run_batch(
        self,
        batch: Iterable[PipelineInput[TextContent]],
        api_key: str = None,
        on_output: Callable[
            [PipelineInput[TextContent], Output[TextContent]], None
        ] = None,
        on_error: Callable[[PipelineInput[TextContent], Exception], None] = None,
        multilingual: bool = False,
    ) -> BatchResponse:
        """
        Runs the pipeline on a batch of input texts.

        ## Parameters

        `batch: Iterable[PipelineInput]`
            The input texts to be processed.
        `api_key: str, optional`
            An API key to be used in this API call. If not provided, `self.api_key` is used.
        `on_output: Callable[[Input, Output], None]`
            Action to perform on successful output, by default creates a dict mapping inputs to outputs
        `on_error: Callable[[Input, Exception], None]`
            Action to perform on error, by default creates a dict mapping inputs to errors

        ## Returns

        Unless on_output/on_error are modified, returns a dictionary mapping inputs to the produced `Output` objects, each containing the results of the Skills in the pipeline.

        ## Raises

        `InputError` if the input is is invalid or is of an incompatible type for the pipeline.
        `APIKeyError` if the API key is invalid, expired, or missing quota.
        `ServerError` if an internal server error occured.
        """
        return _async_run_nested(
            self.run_batch_async(batch, api_key, on_output, on_error, multilingual)
        )

    async def run_batch_async(
        self,
        batch: Iterable[PipelineInput[TextContent]],
        api_key: str = None,
        on_output: Callable[
            [PipelineInput[TextContent], Output[TextContent]], None
        ] = None,
        on_error: Callable[[PipelineInput[TextContent], Exception], None] = None,
        multilingual: bool = False,
    ) -> Awaitable[BatchResponse]:
        """
        Runs the pipeline on a batch of input texts asynchronously.

        ## Parameters

        `batch: Iterable[str | Input]`
            The input texts to be processed.
        `api_key: str, optional`
            An API key to be used in this API call. If not provided, `self.api_key` is used.
        `on_output: Callable[[Input, Output], None]`
            Action to perform on successful output, by default creates a dict mapping inputs to outputs
        `on_error: Callable[[Input, Exception], None]`
            Action to perform on error, by default creates a dict mapping inputs to errors

        ## Returns

        Unless on_output/on_error are modified, returns an Awaitable with a dictionary mapping inputs to the produced `Output` objects, each containing the results of the Skills in the pipeline.

        ## Raises

        `InputError` if the input is is invalid or is of an incompatible type for the pipeline.
        `APIKeyError` if the API key is invalid, expired, or missing quota.
        `ServerError` if an internal server error occured.
        """
        outputs = BatchResponse()
        await process_batch(
            batch,
            self.steps,
            on_output if on_output else outputs.__setitem__,
            on_error if on_error else outputs.__setitem__,
            api_key=api_key or self.api_key or oneai.api_key,
            multilingual=multilingual or self.multilingual or oneai.multilingual,
        )
        return outputs

    def __repr__(self) -> str:
        return f"oneai.Pipeline({self.steps})"


# for jupyter environment, to avoid "asyncio.run() cannot be called from a running event loop"
pool = concurrent.futures.ThreadPoolExecutor()
is_36 = sys.version_info[:2] == (3, 6)

if os.name == "nt" and not is_36:  # Windows
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


def _async_run_nested(coru):
    if is_36:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(coru)
    else:
        try:
            asyncio.get_running_loop()
            return pool.submit(asyncio.run, coru).result()
        except RuntimeError:
            pass

        return asyncio.run(coru)



from dataclasses import replace
from typing import List

import oneai
from oneai.classes import Label, Labels, Output, Skill, Utterance, TextContent


def build_output(
    skills: List[Skill],
    raw_output: dict,
) -> Output:
    if oneai.DEBUG_RAW_RESPONSES:
        return raw_output

    def get_text(index) -> TextContent:
        # get the input text for this Output object. use index=-1 to get the original input text
        # text can be returned as a simple str or parsed to match a given input type
        text = (
            raw_output["output"][index]["contents"]
            if index >= 0
            else raw_output["input"]
        )

        if not text:
            return ""
        if len(text) > 1 or (text and "speaker" in text[0]):
            return [Utterance.from_dict(u) for u in text]
        return text[0]["utterance"]

    def split_pipeline(skills: List[Skill], i: int):
        # split pipeline at a generator Skill
        first, second = skills[: i + 1], skills[i + 1 :]
        if skills[i].labels_attr:
            # handle skills that create both text and labels
            clone = replace(skills[i], text_attr=None)
            second = (clone, *second)
        return first, second

    def build_internal(output_index: int, skills: List[Skill]) -> "Output":
        text = get_text(output_index)
        # temporary fix- if 1st skill is not a generator, use input_text, not output[0].text,
        # since output[0].text is corrupted (not parsable) for conversation inputs
        output_index = max(output_index, 0)
        labels = [
            Label.from_dict(label)
            for label in raw_output["output"][output_index]["labels"]
        ]
        data = []
        for i, skill in enumerate(skills):
            if skill.text_attr:
                skills, next_skills = split_pipeline(skills, i)
                data.append(build_internal(output_index + 1, next_skills))
                break
            else:
                data.append(
                    Labels(filter(lambda label: label.skill == skill.api_name, labels))
                )
        return Output(text=text, skills=list(skills), data=data)

    generator = raw_output["output"][0].get("text_generated_by_step_id", 0) - 1
    if generator < 0:
        return build_internal(-1, skills)
    else:
        # edge case- first Skill is a generator, or a generator preceded by Skills that didn't generate output
        # in this case the API will skip these Skills,
        # so we need to create filler objects to match the expected structure
        skills, next_skills = split_pipeline(skills, generator)
        return Output(
            text=get_text(-1),
            skills=list(skills),
            data=[Labels()] * generator + [build_internal(0, next_skills)],
        )



__version__ = "0.9.16"
__package__ = "oneai"

import logging
import oneai.logger
from typing_extensions import Final
import oneai.skills as skills
from oneai.classes import *
from oneai.pipeline import Pipeline
import oneai.clustering as clustering
import oneai.parsing as parsing
import oneai.util as util

URL: Final[str] = "https://api.oneai.com"
"""
Base URL for the pipeline API. Only change if you know what you're doing.
"""

api_key = None
"""
Default api key to use. You can use different keys for different pipelines by setting the `Pipeline.api_key` attribute or passing an `api_key` parameter to `Pipeline.run` calls.
"""

multilingual = False
"""
Whether to allow multilingual inputs by default. You can use different values for different pipelines by setting the `Pipeline.multilingual` attribute or passing a `multilingual` parameter to `Pipeline.run` calls.
"""

MAX_CONCURRENT_REQUESTS: Final[int] = 2
"""
Max number of allowed concurrent requests to be made by the SDK.
Currently only enforced on `pipeline.run_batch`, other calls may be limited by the API
"""
DEBUG_RAW_RESPONSES = False
"""
Debug flag, return raw API responses instead of structured `Output` object. Only enable if you know what you're doing
"""
DEBUG_LOG_REQUESTS = False
"""
Debug flag, log all requests made by the SDK.
"""

logger: logging.Logger = logging.getLogger("oneai")



from datetime import datetime, timedelta
from dateutil import parser as dateutil
import io
import os
from base64 import b64encode
import validators
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    TextIO,
    Tuple,
    TypeVar,
    Union,
    Hashable,
)
from typing_extensions import dataclass_transform
from warnings import warn


from oneai.exceptions import InputError

if TYPE_CHECKING:
    from oneai.skills import OutputAttrs


@dataclass
class Utterance:
    speaker: str
    utterance: str
    timestamp: Optional[timedelta] = None

    @classmethod
    def from_dict(cls, u: Dict[str, str]) -> "Utterance":
        return cls(
            u["speaker"],
            u["utterance"],
            timestamp_to_timedelta(u.get("timestamp", None)),
        )

    def __repr__(self) -> str:
        return (
            f"\n\t{self.timestamp} {self.speaker}: {self.utterance}"
            if self.timestamp
            else f"\n\t{self.speaker}: {self.utterance}"
        )


TextContent = TypeVar("TextContent", bound=Union[str, List["Utterance"]])
PipelineInput = Union["Input[TextContent]", TextContent, TextIO, BinaryIO]

# extension -> content_type, input_type
CONTENT_TYPES: Dict[str, Tuple[str, str]] = {
    ".json": ("application/json", "conversation"),
    ".txt": ("text/plain", "article"),
    ".srt": ("text/plain", "conversation"),
    ".wav": ("audio/wav", "conversation"),
    ".mp3": ("audio/mpeg", "conversation"),
    ".mp4": ("audio/mpeg", "conversation"),
    ".html": ("text/plain", "article"),
    ".pdf": ("text/pdf", "article"),
}


@dataclass(frozen=True)
class Skill:
    """
    A base class for all Language Skills. Use predefined subclasses of this class, or use this class to define your own Skills.

    A Language Skill is a package of trained NLP models. Skills accept text and respond with processed texts and extracted metadata.

    Process texts with Skills using `Pipeline`s

    ### Skill types
    * Generator Skills (`text_attr is not None`) process the input and produce a new text based on it. Examples include `Summarize`, `Proofread`.
    * Analyzer Skills (`text_attr is None`) scan the input and extract structured data. Examples include `Emotions`, `Topics`.

    ## Attributes

    `api_name: str`
        The name of the Skill in the pipeline API.
    `text_attr: str`
        The attribute name of the Skill's output text in the Output object (Generator Skills only).
    `labels_attr: str`
        The attribute name of the Skill's output labels in the Output object.
    `params: dict[str, Any]`
        The parameters of the Skill. See the documentation for each Skill for a list of available parameters.
    """

    api_name: str = ""
    text_attr: Optional[str] = None
    labels_attr: Optional[str] = None
    params: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        # backwards compatibility
        if self.labels_attr is None and self.text_attr is None:
            object.__setattr__(self, "labels_attr", self.api_name)

    def asdict(self) -> dict:
        return {
            "skill": self.api_name,
            "params": {k: v for k, v in self.params.items() if v is not None},
        }


@dataclass_transform()
def skillclass(
    api_name: str = "",
    text_attr: Optional[str] = None,
    labels_attr: Optional[str] = None,
):
    """
    A decorator for defining a Language Skill class. Decorate subclasses of `Skill` with this to provide default values for instance attributes.

    ## Example

    >>> @skillclass(api_name="my-skill", text_attr="my_result")
    ... class MySkill(Skill):
    ...     ratio: float = 0.2
    >>> s = Summarize(ratio=0.5)
    >>> s.ratio
    0.5
    >>> pipeline = Pipeline([s])
    >>> output = pipeline.run("Text to be processed with MySkill")
    >>> output.my_result
    "Result text, processed with MySkill"
    """

    def wrap(cls):
        if not issubclass(cls, Skill):
            warn(
                f"warning: class {cls.__name__} decorated with @skillclass does not inherit Skill",
                stacklevel=2,
            )

        # remove class variables
        classVars = {
            k: getattr(cls, k, None)
            for k in cls.__annotations__
            if k not in Skill.__annotations__
        }
        for k in classVars:
            if hasattr(cls, k):
                delattr(cls, k)

        def __init__(self, **params):
            Skill.__init__(
                self,
                api_name=api_name,
                text_attr=text_attr,
                labels_attr=labels_attr,
                params=params,
            )

            for k, v in classVars.items():
                if k not in params:
                    setattr(self, k, v)

        def __getattr__(self, name):
            return self.params[name]

        def __setattr__(self, name, value):
            if "params" not in self.__dict__:
                return object.__setattr__(self, name, value)
            self.params[name] = value

        cls.__init__ = __init__
        cls.__getattr__ = __getattr__
        cls.__setattr__ = __setattr__
        return cls

    return wrap


class Input(Generic[TextContent]):
    """
    A base class for all input texts, allowing structured representations of inputs.

    ## Attributes

    `text: TextContent`
        Input text. Either `str` or `list[Utterance]` (conversation).
    `type: str`
        A type hint for the API, suggesting which models to use when processing the input.
    `content_type: str`
        The content type of the input.
    `encoding: str`
        The encoding of the input.
    `metadata: dict`
        Optional metadata to be associated with the input in clustering collections.
    `datetime: datetime`
        Optional datetime to be associated with the input in clustering collections.
    """

    def __init__(
        self,
        text: TextContent,
        *,
        type: str = None,
        content_type: str = None,
        encoding: str = None,
        metadata: Dict[str, any] = None,
        datetime: datetime = None,
    ):
        self.text: TextContent = text
        self.type = type
        self.content_type = content_type
        self.encoding = encoding
        self.metadata = metadata
        self.datetime = datetime

    @classmethod
    def wrap(
        cls, text: PipelineInput[TextContent], sync: bool = True
    ) -> "Input[TextContent]":
        if isinstance(text, cls):
            return text
        elif isinstance(text, str):
            if validators.url(text):
                return cls(text, type="article", content_type="text/uri-list")
            else:
                return cls(text, type="article", content_type="text/plain")
        elif isinstance(text, list) and (
            len(text) == 0 or isinstance(text[0], Utterance)
        ):
            return cls(text, type="conversation", content_type="application/json")
        elif isinstance(text, io.IOBase):
            name, mode = text.name, text.mode
            _, ext = os.path.splitext(name)
            if ext not in CONTENT_TYPES:
                raise InputError(
                    message=f"unsupported file extension {ext}",
                    details="see supported files in docs",
                )
            content_type, input_type = CONTENT_TYPES[ext]
            if "b" not in mode:
                return cls(text.read(), type=input_type, content_type=content_type)
            elif sync:
                data = b64encode(text.read()).decode("ascii")
                return cls(
                    data, type=input_type, content_type=content_type, encoding="base64"
                )
            else:
                return cls(text, type=input_type, content_type=content_type)
        else:
            raise ValueError(f"invalid content type {type(text)}")


def timestamp_to_timedelta(timestamp: str) -> timedelta:
    if not timestamp:
        return None
    dt = dateutil.parse(timestamp)
    return timedelta(
        hours=dt.hour, minutes=dt.minute, seconds=dt.second, microseconds=dt.microsecond
    )


@dataclass
class Span:
    start: int
    end: int
    section: int = 0
    text: str = None

    @classmethod
    def from_dict(cls, objects: List[dict], text: str) -> "List[Span]":
        return (
            []
            if not objects
            else [
                cls(
                    start=object.get("start", None),
                    end=object.get("end", None),
                    section=object.get("section", None),
                    text=text,
                )
                for object in objects
            ]
        )


@dataclass
class Label:
    """
    Represents a label, marking a part of the input text. Attribute values largely depend on the Skill the labels were produced by.

    ## Attributes

    `type: str`
        Label type, e.g. 'entity', 'topic', 'emotion'.
    `skill: str`
        The name of the Skill that produced the label.
    `name: str`
        Label class name, e.g. 'PERSON', 'happiness', 'POS'.
    `output_spans: list[Span]`
        The spans in the output text that are marked with the label.
    `input_spans: list[Span]`
        The spans in the input text that are relevant to the label. Only appears if the label was produced by a Skill that supports input spans.
    `span_text: str`
        The text of the label.
    `timestamp: str`
        For audio inputs, the timestamp of the start of the label.
    `timestamp_end: str`
        For audio inputs, the timestamp of the end of the label.
    `value: str`
        The value of the label.
    `data: Dict[str, Any]`
        Additional data associated with the label.
    """

    type: str = ""
    skill: str = ""
    name: str = ""
    _span: List[int] = field(default_factory=lambda: [0, 0], repr=False)
    output_spans: List[Span] = field(default_factory=list)
    input_spans: List[Span] = field(default_factory=list)
    span_text: str = ""
    timestamp: timedelta = None
    timestamp_end: timedelta = None
    value: str = ""
    data: dict = field(default_factory=dict)

    @property
    def span(self) -> Span:
        warn(
            "`Label.span` is deprecated, use `Label.output_spans` instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._span

    @classmethod
    def from_dict(cls, object: dict) -> "Label":
        return cls(
            type=object.pop("type", ""),
            skill=object.pop("skill", ""),
            name=object.pop("name", ""),
            output_spans=Span.from_dict(
                object.pop("output_spans", []), object.get("span_text", None)
            ),
            input_spans=Span.from_dict(
                object.pop("input_spans", []), object.get("span_text", None)
            ),
            _span=object.pop("span", [0, 0]),
            span_text=object.pop("span_text", ""),
            value=object.pop("value", ""),
            data=object.pop("data", {}),
            timestamp=timestamp_to_timedelta(object.pop("timestamp", "")),
            timestamp_end=timestamp_to_timedelta(object.pop("timestamp_end", "")),
        )

    def __repr__(self) -> str:
        return (
            "oneai.Label("
            + ", ".join(
                f"{k}={repr(v)}"
                for k, v in self.__dict__.items()
                if v and not k.startswith("_")
            )
            + ")"
        )


class Labels(List[Label]):
    """
    Wrapper object for a list of `Label` objects. Provides convenience methods to query labels by attribute.

    ## Properties

    `values: list[str]`
        A list of all values of the labels.
    `names: list[str]`
        A list of all names of the labels.
    `input_spans: list[list[Span]]`
        A list of all input spans of the labels.
    `output_spans: list[list[Span]]`
        A list of all output spans of the labels.
    `span_texts: list[str]`
        A list of all span texts of the labels.
    """

    @property
    def values(self) -> List[Any]:
        return [l.value for l in self]

    @property
    def names(self) -> List[str]:
        return [l.name for l in self]

    @property
    def input_spans(self) -> List[List[Span]]:
        return [l.input_spans for l in self]

    @property
    def output_spans(self) -> List[List[Span]]:
        return [l.output_spans for l in self]

    @property
    def span_texts(self) -> List[str]:
        return [l.span_text for l in self]


class Output(Input[TextContent], OutputAttrs if TYPE_CHECKING else object):
    """
    Represents the output of a pipeline. The structure of the output is dynamic, and corresponds to the Skills used and their order in the pipeline.
    Skill outputs can be accessed as attributes, either with the `api_name` of the corresponding Skill or the `output_attr` field.

    ## Attributes

    `text: TextContent`
        The input text from which this `Output` instance was produced.
    `skills: List[Skill]`
        The Skills used to process `text` and produce this `Output` instance.
    See `OutputAttrs` for the attributes generated by different Skills.
    """

    def __init__(
        self,
        text: TextContent,
        skills: List[Skill] = [],
        data: List[Union[Labels, "Output"]] = [],
    ):
        super().__init__(
            text,
            type="article" if isinstance(text, str) else "conversation",
            content_type="text/plain" if isinstance(text, str) else "application/json",
        )

        self.skills = skills
        for skill, value in zip(skills, data):
            setattr(self, skill.text_attr or skill.labels_attr or skill.api_name, value)

    def __dir__(self) -> Iterable[str]:
        return super().__dir__() + [
            skill.text_attr or skill.labels_attr or skill.api_name
            for skill in self.skills
        ]

    def __repr__(self) -> str:
        result = f"oneai.Output(text={repr(self.text)}"
        for skill in self.skills:
            attr = skill.text_attr or skill.labels_attr or skill.api_name
            result += f", {attr}={repr(getattr(self, attr))}"
        return result + ")"


class BatchResponse:
    def __init__(self):
        self._data: Dict[Input, Output] = {}

    def __setitem__(self, key: Input, value: Output):
        self._data[key] = value

    def __getitem__(self, key: Input) -> Output:
        return (
            self._data[key]
            if isinstance(key, Hashable) and key in self._data
            else next(v for k, v in self._data.items() if k.text == key)
        )



import json
from typing import Dict, Union
from aiohttp import ClientResponse

# todo: input type validation errors


class OneAIError(Exception):
    """
    A base class for all errors raised by the API.

    ## Attributes

    `status_code: int`
        The API status code of the error.
    `message: str`
        A human-readable message describing the error.
    `details: str`
        A string containing details about the error.
    """

    def __init__(
        self,
        status_code: int = 0,
        message: str = "",
        details: str = "",
        request_id: str = "",
    ):
        self.status_code = status_code
        self.message = message
        self.details = details
        self.request_id = request_id

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}("
            + (f"status_code={self.status_code}, " if self.status_code else "")
            + (f"message={self.message}" if self.message else "")
            + (f", details={self.details}" if self.details else "")
            + (f", request_id={self.request_id})" if self.request_id else ")")
        )

    __repr__ = __str__


class InputError(OneAIError):
    """An error raised when the input is invalid or is of an incompatible type for the pipeline."""


class APIKeyError(OneAIError):
    """An error raised when the API key is invalid, expired, or missing quota."""


class ServerError(OneAIError):
    """An error raised when the an internal server error occured."""


errors = {  # map http status codes to OneAIError subclasses
    400: InputError,
    401: APIKeyError,
    403: APIKeyError,
    500: ServerError,
    503: ServerError,
}


async def handle_unsuccessful_response(response: Union[ClientResponse, Dict]):
    status, reason = 0, ""
    if isinstance(response, ClientResponse):
        try:
            status, reason = response.status, response.reason
            response = json.loads(await response.content.read())
        except:
            response = {}
    else:
        status = int(str(response.get("status_code", 0))[:3])
    raise errors.get(status, ServerError)(
        response.get("status_code", status),
        response.get("message", reason),
        response.get("details", ""),
        response.get("request_id", ""),
    )


def validate_api_key(api_key: str):
    if api_key is None or not api_key:
        raise APIKeyError(
            60001,
            "Missing API key",
            "Please provide a valid API key, either by setting the global `oneai.api_key` or passing the `api_key` parameter",
        )



import logging

PREFIX = "\33[34m●\33[36m▲\33[35m▮\33[0m"

notebook = False
try:
    if get_ipython().__class__.__name__ == "ZMQInteractiveShell":
        notebook = True
except NameError:
    pass


class Formatter(logging.Formatter):
    def format(self, record):
        if record.levelno == logging.DEBUG:
            if notebook:
                self._style._fmt = f"{PREFIX} %(message)s\r"
            else:
                self._style._fmt = f"\033[K{PREFIX} %(message)s\033[F"
        if record.levelno == logging.DEBUG + 1:
            self._style._fmt = ""
        elif record.levelno == logging.WARNING:
            self._style._fmt = "\33[33m%(message)s\33[0m"
        elif record.levelno == logging.ERROR:
            self._style._fmt = "\33[91m%(message)s\33[0m"
        return super().format(record)


formatter = Formatter()
handler = logging.StreamHandler()
handler.setFormatter(formatter)

logger = logging.getLogger("oneai")
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)



from datetime import datetime
import json
import requests
import urllib.parse
from typing import Union, Callable, Any
from typing_extensions import Literal
import oneai, oneai.api


API_DATE_FORMAT = "%Y-%m-%d"

ENDPOINT = "clustering/v1/collections"


def build_query_params(
    sort: Literal["ASC", "DESC"] = None,
    limit: int = None,
    from_date: Union[datetime, str] = None,
    to_date: Union[datetime, str] = None,
    date_format: str = API_DATE_FORMAT,
    item_metadata: str = None,
):
    if from_date:
        if isinstance(from_date, str):
            from_date = datetime.strptime(from_date, date_format)
        from_date = from_date.strftime(API_DATE_FORMAT)

    if to_date:
        if isinstance(to_date, str):
            to_date = datetime.strptime(to_date, date_format)
        to_date = to_date.strftime(API_DATE_FORMAT)

    params = {
        "sort": sort,
        "limit": limit if limit else 20,
        "from-date": from_date,
        "to-date": to_date,
        "include-phrases": False,
        "item-metadata": item_metadata,
    }
    return urllib.parse.urlencode({k: v for k, v in params.items() if v})


def get_clustering_paginated(
    path: str,
    api_key: str,
    result_key: str,
    parent: Any,
    from_dict: Callable[[Any, dict], Any],
    sort: Literal["ASC", "DESC"] = None,
    limit: int = None,
    from_date: Union[datetime, str] = None,
    to_date: Union[datetime, str] = None,
    date_format: str = API_DATE_FORMAT,
    item_metadata: str = None,
):
    path_query = f"{path}?{build_query_params(sort, limit, from_date, to_date, date_format, item_metadata)}"
    page = 0
    counter = 0
    results = [None]

    while results and ((not limit) or counter < limit):
        response = get_clustering(f"{path_query}&page={page}", api_key)
        results = [
            (from_dict(parent, result) if parent else from_dict(result))
            for result in response[result_key]
        ]
        yield from results
        counter += len(results)
        page += 1

        if page >= response.get("total_pages", 0):
            break


def get_clustering(path: str, api_key: str = None):
    api_key = api_key or oneai.api_key
    if not api_key:
        raise Exception("API key is required")
    headers = {
        "api-key": api_key,
        "Content-Type": "application/json",
        "User-Agent": f"python-sdk/{oneai.__version__}/{oneai.api.uuid}",
    }
    if oneai.DEBUG_LOG_REQUESTS:
        oneai.logger.debug(f"GET {oneai.URL}/{ENDPOINT}/{path}\n")
        oneai.logger.debug(f"headers={json.dumps(headers, indent=4)}\n")
    response = requests.get(
        f"{oneai.URL}/{ENDPOINT}/{path}",
        headers=headers,
    )
    return json.loads(response.content)


def post_clustering(path: str, data: dict, api_key: str = None):
    api_key = api_key or oneai.api_key
    if not api_key:
        raise Exception("API key is required")
    headers = {
        "api-key": api_key,
        "Content-Type": "application/json",
        "User-Agent": f"python-sdk/{oneai.__version__}/{oneai.api.uuid}",
    }
    if oneai.DEBUG_LOG_REQUESTS:
        oneai.logger.debug(f"POST {oneai.URL}/{ENDPOINT}/{path}\n")
        oneai.logger.debug(f"headers={json.dumps(headers, indent=4)}\n")
        oneai.logger.debug(f"data={json.dumps(data, indent=4)}\n")
    response = requests.post(
        f"{oneai.URL}/{ENDPOINT}/{path}", headers=headers, json=data
    )
    return json.loads(response.content)



import re
from typing import List
import oneai
from oneai.classes import Utterance


# v 1.6.1
def parse_conversation(text: str, strict=False) -> List[Utterance]:
    """
    Parse a string with a conversation format into a structured `Utterance` list representing the conversation.

    ## Parameters

    `text: str`
        The text to parse.

    ## Returns

    A list of `Utterance` objects produced from `text`.

    ## Raises

    `ValueError` if `text` is not in a valid conversation format.
    """

    srt_regex = re.compile(
        r"\d+\n\d{1,2}:\d{2}:\d{2}[,.]\d{1,3} --> \d{1,2}:\d{2}:\d{2}[,.]\d{1,3}"
    )
    match = srt_regex.match(text)
    if match:
        data_array = srt_regex.split(text)
        return [
            Utterance(speaker="SPEAKER", utterance=line.strip().replace("\n", " "))
            for line in data_array[1:]
        ]

    result = []
    lines = re.split(r"\r?\n", text.strip())
    firstLine = True
    structure = None
    currentLineInfo = None
    waitForTextLine = False
    # weak = False
    previousObject = None

    for i, line in enumerate(lines):
        if _isEmptyOrWhitespace(line):
            continue

        if waitForTextLine:
            previousObject["text"] = line.strip()
            # previousObject["text_line"] = i
            waitForTextLine = False
            continue

        currentLineInfo = _parseSpeakerLine(line)
        if currentLineInfo is None:
            if firstLine:
                raise ValueError(f"Invalid conversation format at line {i}")
            previousObject["text"] += "\n" + line.strip()
            # weak = True
            continue

        if firstLine:
            structure = currentLineInfo

        # weak |= currentLineInfo["weak"]
        if strict and not _comp(structure, currentLineInfo):
            raise ValueError(
                f"Differing conversation format at line {i}, run with strict=False to ignore"
            )

        firstLine = False

        previousObject = {
            "speaker": currentLineInfo["speaker"],
            "text": currentLineInfo["text"],
            # "speaker_line": i,  # what are these properties for? do I want them in Utterance objects?
            # "text_line": i,
            # "speaker_length": currentLineInfo["speaker_end"],
        }
        if currentLineInfo["timestamp"]:
            previousObject["timestamp"] = currentLineInfo["timestamp"]

        result.append(previousObject)
        waitForTextLine = not currentLineInfo["hasText"]
    if previousObject and _isEmptyOrWhitespace(previousObject["text"]):
        result.pop()

    return [
        Utterance(
            speaker=u["speaker"],
            utterance=u["text"],
            timestamp=u.get("timestamp", None),
        )
        for u in result
    ]


def _isEmptyOrWhitespace(text):
    return (not text) or (text.isspace())


def _parseSpeakerLine(text: str):
    value = {
        "weak": True,
        "preTime": False,
        "speaker": None,
        "speaker_end": None,
        "time": False,
        "timestamp": None,
        "timestamp_full_match_string": None,
        "separator": False,
        "hasText": False,
        "text": None,
    }

    ################################################
    # extracting timestamp from text
    matchArea = text[:40]
    colonPos = matchArea.find(":")
    timestampFound = get_timestamp(matchArea, value)
    signatureEndPos = 0
    if timestampFound:
        if colonPos != -1 and colonPos < value["timestamp_position"]:
            timestampFound = False
            value["time"] = False
            value["timestamp"] = None
        else:
            if value["timestamp_position"] == 0:
                value["preTime"] = True
            text = text.replace(value["timestamp_full_match_string"], "")
            matchArea = matchArea.replace(value["timestamp_full_match_string"], "")
            signatureEndPos = len(value["timestamp_full_match_string"])
    ################################################
    # check if speaker only, in all caps - WEAK PATTERN
    match = (
        re.search(r"^[ A-Za-z_-]{3,20}$", text)
        if timestampFound
        else re.search(r"^[ A-Z_-]{3,20}$", text)
    )
    if match is not None:
        value["weak"] = not timestampFound
        value["speaker"] = match[0].strip()
        # end position for speaker signature area (for highlighting),
        # use match[0].length to include whitespace
        value["speaker_end"] = len(match[0]) + signatureEndPos
        value["hasText"] = False
        return value

    # update colon position after timestamp removal
    if timestampFound:
        colonPos = matchArea.find(":")

    if colonPos == -1 and not timestampFound:
        return None  # failed to find signature

    if colonPos == -1:  # only timestamp
        if len(text) != 0:
            return None  # if text after timestamp, fail
        value["weak"] = True
        value["speaker"] = "Speaker"
        value["speaker_end"] = signatureEndPos
        value["hasText"] = False

    value["separator"] = True

    # if no whitespace after speaker, fail same line text
    textPos = colonPos + 1
    value["hasText"] = textPos < len(text.rstrip()) - 1
    if value["hasText"] and " \t\n\r\v".find(text[textPos]) == -1:
        return None

    value["weak"] = False
    value["text"] = text[textPos:].strip() if value["hasText"] else None
    value["speaker"] = text[:colonPos].strip()
    value["speaker_end"] = signatureEndPos + colonPos
    return value


def _comp(a, b):
    return (
        a["separator"] == b["separator"]
        and a["time"] == b["time"]
        and a["preTime"] == b["preTime"]
        and a["hasText"] == b["hasText"]
    )


def get_timestamp(text, value):
    # match preceding timestamp "[3:07 PM, 3/15/2022] Adam Hanft: Helps"
    match = re.search(r"(^\s*\[?\s*)([0-9:,\sPAM/]{4,23})(\]?)\s*", text)
    if match is not None and (match[3] or match[0].find("/") != -1):
        value["preTime"] = True
        value["weak"] = False
        value["time"] = True
        value["timestamp"] = match[2].strip()
        value["timestamp_position"] = index_of_group(match, 2)
        value["timestamp_full_match_string"] = match[0]
        return True

    #                  optinal      [        timestamp                 ]  \s*
    match = re.search(r"(^\s*)?(\[?)(\d{1,2}:\d{1,2})(:\d{1,2})?(\.\d*)?(\]?\s*)", text)
    if match is not None:
        value["weak"] = False
        value["time"] = True
        value["timestamp_position"] = match.start()
        value["timestamp_full_match_string"] = match[0]
        # capture timestamp without non-captured groups
        value["timestamp"] = text[
            index_of_group(match, 3) : index_of_group(match, 6)
        ].strip()
        return True

    return False


def index_of_group(match: re.Match, group: int):
    ix = match.start()
    for i in range(1, group):
        if match[i]:
            ix += len(match[i])
    return ix



from dataclasses import dataclass, field
from datetime import datetime
import urllib.parse
from typing import Generator, List, Union
from typing_extensions import Literal
import oneai
from oneai.api import get_clustering, get_clustering_paginated, post_clustering
from oneai.classes import Input, PipelineInput

API_DATE_FORMAT = "%Y-%m-%d"


def get_collections(
    api_key: str = None,
    *,
    limit: int = None,
) -> Generator["Collection", None, None]:
    yield from get_clustering_paginated(
        "",
        api_key,
        "collections",
        None,
        from_dict=Collection,
        limit=limit,
    )


@dataclass
class Item(Input[str]):
    id: int
    text: str
    datetime: datetime
    distance: float
    phrase: "Phrase" = field(repr=False)
    cluster: "Cluster" = field(repr=False)
    metadata: dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, phrase: "Phrase", object: dict) -> "Item":
        item = cls(
            id=object["id"],
            text=object["original_text"],
            datetime=datetime.strptime(object["create_date"], f"%Y-%m-%d %H:%M:%S.%f"),
            distance=object["distance_to_phrase"],
            phrase=phrase,
            cluster=phrase.cluster,
            metadata={k: v[0]["value"] for k, v in object["metadata"].items()},
        )
        item.type = "article"
        item.content_type = "text/plain"
        return item


@dataclass
class Phrase:
    id: int
    text: str
    item_count: int
    cluster: "Cluster" = field(repr=False)
    collection: "Collection" = field(repr=False)
    metadata: dict = field(default_factory=dict)

    def get_items(
        self,
        *,
        limit: int = None,
        item_metadata: str = None,
    ) -> List[Item]:
        yield from get_clustering_paginated(
            f"{self.collection.id}/phrases/{self.id}/items",
            self.collection.api_key,
            "items",
            self,
            Item.from_dict,
            limit=limit,
            item_metadata=item_metadata,
        )

    @classmethod
    def from_dict(cls, cluster: "Cluster", object: dict) -> "Phrase":
        return cls(
            id=int(object["phrase_id"]),
            text=object.get("text", ""),
            item_count=object["items_count"],
            cluster=cluster,
            collection=cluster.collection,
            metadata=object["metadata"],
        )


@dataclass
class Cluster:
    id: int
    text: str
    phrase_count: int
    item_count: int
    collection: "Collection" = field(repr=False)
    metadata: dict = field(default_factory=dict)

    def get_phrases(
        self,
        *,
        sort: Literal["ASC", "DESC"] = "ASC",
        limit: int = None,
        from_date: Union[datetime, str] = None,
        to_date: Union[datetime, str] = None,
        date_format: str = API_DATE_FORMAT,
        item_metadata: str = None,
    ) -> Generator[Phrase, None, None]:
        yield from get_clustering_paginated(
            f"{self.collection.id}/clusters/{self.id}/phrases",
            self.collection.api_key,
            "phrases",
            self,
            Phrase.from_dict,
            sort,
            limit,
            from_date,
            to_date,
            date_format,
            item_metadata,
        )

    def get_items(
        self,
        *,
        sort: Literal["ASC", "DESC"] = "ASC",
        limit: int = None,
        from_date: Union[datetime, str] = None,
        to_date: Union[datetime, str] = None,
        date_format: str = API_DATE_FORMAT,
        item_metadata: str = None,
    ) -> Generator[Phrase, None, None]:
        yield from get_clustering_paginated(
            f"{self.collection.id}/clusters/{self.id}/items",
            self.collection.api_key,
            "items",
            self,
            Item.from_dict,
            sort,
            limit,
            from_date,
            to_date,
            date_format,
            item_metadata,
        )

    def add_items(self, items: List[PipelineInput[str]]):
        url = f"{self.collection.id}/items"
        data = [
            {
                "text": item.text if isinstance(item, Input) else item,
                "item_metadata": item.metadata if isinstance(item, Input) else None,
                "force-cluster-id": self.id,
            }
            for item in items
        ]
        post_clustering(url, data, self.collection.api_key)

    @classmethod
    def from_dict(cls, collection: "Collection", object: dict) -> "Cluster":
        return cls(
            id=int(object["cluster_id"]),
            text=object["cluster_phrase"],
            phrase_count=object["phrases_count"],
            item_count=object["items_count"],
            collection=collection,
            metadata=object["metadata"],
        )


class Collection:
    def __init__(self, id: str, api_key: str = None):
        self.id = id
        self.api_key = api_key or oneai.api_key

    def get_clusters(
        self,
        *,
        sort: Literal["ASC", "DESC"] = "ASC",
        limit: int = None,
        from_date: Union[datetime, str] = None,
        to_date: Union[datetime, str] = None,
        date_format: str = API_DATE_FORMAT,
        item_metadata: str = None,
    ) -> Generator[Cluster, None, None]:
        yield from get_clustering_paginated(
            f"{self.id}/clusters",
            self.api_key,
            "clusters",
            self,
            Cluster.from_dict,
            sort,
            limit,
            from_date,
            to_date,
            date_format,
            item_metadata,
        )

    def find(self, query: str, threshold: float = 0.5) -> List[Cluster]:
        params = {
            "text": query,
            "similarity-threshold": threshold,
        }

        url = f"{self.id}/clusters/find?{urllib.parse.urlencode(params)}"
        # return [
        #     Cluster.from_dict(self, cluster)
        #     for cluster in get_clustering(url, self.api_key)
        # ]
        return get_clustering(url, self.api_key)

    def add_items(
        self, items: List[PipelineInput[str]], force_new_clusters: bool = False
    ):
        def build_item(input: PipelineInput[str]):
            result = {
                "text": input.text if isinstance(input, Input) else str(input),
                "force-new-cluster": force_new_clusters,
            }
            if hasattr(input, "metadata") and input.metadata:
                result["item_metadata"] = input.metadata
            if hasattr(input, "datetime") and input.datetime:
                result["timestamp"] = int(input.datetime.timestamp())
            return result

        url = f"{self.id}/items"
        data = [build_item(item) for item in items]
        print(post_clustering(url, data, self.api_key))

    def __repr__(self) -> str:
        return f"oneai.Collection({self.id})"



from typing import List
from typing_extensions import Literal

from oneai.classes import Labels, Skill, Utterance, Output, skillclass


@skillclass(
    api_name="enhance",
    text_attr="proofread",
    labels_attr="replacements",
)
class Proofread(Skill):
    """
    Enhances conversations, removing fillers and mistakes. Only works with `Conversation` inputs

    ## Output Attributes

    `proofread: Output`
        An `Output` object containing the enhanced conversation, and output of the following `Skill`s
    `proofread.text: str`
        The enhanced conversation
    `proofread.replacements: list[Label]`
        A list of `Label` objects, representing changes made to the text

    ## Example

    >>> conversation = [oneai.Utterance(...), ...]
    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Proofread()
    ... ])
    >>> output = pipeline.run(conversation)
    >>> output.proofread
    oneai.Output(text='PROOFREAD TRANSCRIPTION', replacements=[...])
    """


@skillclass(
    api_name="summarize",
    text_attr="summary",
    labels_attr="origins",
)
class Summarize(Skill):
    """
    Provides a summary of the input

    ##  Attributes

    `min_length: int`
        Minimal desired length (words) of the summary. Omit for automatic length
    `max_length: int`
        Maximal desired length (words) of the summary. Omit for automatic length
    `find_origins: bool, default=False`
        Whether to generate origins of the summary, pointing to the input text

    ## Output Attributes

    `summary: Output`
        An `Output` object containing the generated summary, and output of the following `Skill`s
    `summary.text: str`
        The generated summary
    `summary.origins: list[Label]`
        A list of `Label` objects, mapping words from the summary to the corresponding words in the input text. Only generated if `find_origins=True`

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Summarize(min_length=10, find_origins=True)
    ... ])
    >>> output = pipeline.run('YOUR-TEXT')
    >>> output.summary
    oneai.Output(text='SUMMARY', origins=[...])
    """

    min_length: int = None
    max_length: int = None
    find_origins: bool = False


@skillclass(api_name="emotions")
class Emotions(Skill):
    """
    Detects emotions in the input. Supported emotions: [`happiness`, `sadness`, `fear`, `surprise`, `anger`]

    ## Output Attributes

    `emotions: list[Label]`
        A list of `Label` objects, with spans of the input that were detected as conveying an emotion

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Emotions()
    ... ])
    >>> output = pipeline.run("I'm not interested. Don't call me again!")
    >>> output.emotions
    [oneai.Label(type=emotion, span=[20, 40], name=anger)]
    """


@skillclass(api_name="keywords")
class Keywords(Skill):
    """
    Detects keywords in the input

    ## Output Attributes

    `keywords: list[Label]`
        A list of `Label` objects, with detected keywords

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Keywords()
    ... ])
    >>> output = pipeline.run('I love this app, it helps me to organize my tasks')
    >>> output.keywords
    [oneai.Label(type=keyword, span=[12, 15], value=app), oneai.Label(type=keyword, span=[44, 49], value=tasks)]
    """


@skillclass(api_name="highlights")
class Highlights(Skill):
    """
    Detects highlights of the input

    ## Output Attributes

    `highlights: list[Label]`
        A list of `Label` objects, with detected highlights

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Highlights()
    ... ])
    >>> output = pipeline.run('YOUR-TEXT')
    >>> output.keywords
    [oneai.Label(value='A-HIGHLIGHT'), ...]
    """


@skillclass(api_name="sentiments")
class Sentiments(Skill):
    """
    Detects sentiments in the input

    ## Output Attributes

    `sentiments: list[Label]`
        A list of `Label` objects, with spans of the input that were detected as having positive or negative sentiment

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Sentiments()
    ... ])
    >>> output = pipeline.run('I love this app, it helps me to organize my tasks')
    >>> output.sentiments
    [oneai.Label(type=sentiment, span=[0, 49], value=POS)]
    """


@skillclass(api_name="article-topics", labels_attr="topics")
class Topics(Skill):
    """
    Detects topics of the input

    ## Output Attributes

    `topics: list[Label]`
        A list of `Label` objects, with extracted topics

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Topics()
    ... ])
    >>> output = pipeline.run('One AI is an NLP platform for developers')
    >>> output.topics
    [oneai.Label(type=topic, value=Machine Learning), oneai.Label(type=topic, value=Artificial Intelligence), ...]
    """


@skillclass(
    api_name="pdf-extract-text",
    text_attr="pdf_text",
)
class PDFExtractText(Skill):
    """
    Extracts the text from PDF files.

    ## Output

    Main text content from the PDF file
    """


@skillclass(
    api_name="extract-html",
    text_attr="html_article",
)
class HtmlToArticle(Skill):
    """
    Extracts the main text content of an HTML page. Accepts URLs or HTML strings

    ## Output

    Main text content from the HTML page

    ## Example

    >>> resp = requests.get('https://oneai.com/about-us/')
    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.HtmlToArticle()
    ... ])
    >>> output = pipeline.run(resp.text)
    >>> output.html_article.text[:100] + '...'
    One AI - About us
    ABOUT ONE AI
    // Bringing human-level language AI to everyday life, one developer a...
    """


@skillclass(
    api_name="html-extract-text",
    text_attr="html_text",
)
class HtmlAllText(Skill):
    """
    Extracts all text content of an HTML page (including text from control elements). Accepts URLs or HTML strings. Use HTMLExtractArticle to extract the main content only

    ## Output

    Text content from the HTML page

    ## Example

    >>> resp = requests.get('https://oneai.com/about-us/')
    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.HtmlAllText()
    ... ])
    >>> output = pipeline.run(resp.text)
    >>> output.html_text.text[:100] + '...'
    One AI - About usSkip to main contentAPIStudioSkillsIndustriesPricingResearchAboutAbout UsLeadership...
    """


@skillclass(api_name="action-items", labels_attr="action_items")
class ActionItems(Skill):
    """
    ### 'Labs' Skill- this Skill is still in beta and is may produce incorrect results in some cases

    Detects action items in the input

    ## Output Attributes

    `action_items: list[Label]`
        A list of `Label` objects, with detected action items

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.ActionItems()
    ... ])
    >>> output = pipeline.run("It's been a blast talking to you. Let's schedule another meeting for next Sun.")
    >>> output.action_items
    [oneai.Label(type=action-item, span=[34, 78], value=Let's schedule another meeting for next Sun.)]
    """


@skillclass(
    api_name="anonymize",
    text_attr="anonymized",
    labels_attr="anonymizations",
)
class Anonymize(Skill):
    """
    Anonymize texts, removing personal information

    ## Output Attributes

    `anonymized: Output`
        An `Output` object containing the anonymized text, and output of the following `Skill`s
    `anonymized.text: str`
        The anonymized text
    `anonymized.anonymizations: list[Label]`
        A list of `Label` objects, representing anonymized information

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Anonymize()
    ... ])
    >>> output = pipeline.run("hello, I'm Michael, my email is michael@abcde.com")
    >>> output.anonymized.text
    hello, I'm ***, my email is ***
    >>> output.anonymized.anonymizations
    [oneai.Label(type=anonymized, span=[11, 14], value=Michael), oneai.Label(type=anonymized, span=[28, 31], value=michael@abcde.com)]
    """


@skillclass(
    api_name="business-entities",
    text_attr="labs",
    labels_attr="pricing",
)
class Pricing(Skill):
    """
    ### 'Labs' Skill- this Skill is still in beta and is may produce incorrect results in some cases

    Detects pricing and quantities in the input

    ## Output Attributes

    `pricing: list[Label]`
        A list of `Label` objects, with detected pricing and their data

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Pricing()
    ... ])
    >>> output = pipeline.run('it costs 10 USD per barrel')
    >>> output.labs.pricing
    [oneai.Label(name=pricing, data={'amount': 10.0, 'currency': 'USD', 'unit': 'barrel'}, span=[9, 26], value=10 USD per barrel)]
    """


@skillclass(api_name="names")
class Names(Skill):
    """
    Detects and classifies names in the input

    ## Output Attributes

    `names: list[Label]`
        A list of `Label` objects, with detected names and their type

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Names()
    ... ])
    >>> output = pipeline.run('Jim bought shares of Acme Corp.')
    >>> output.names
    [oneai.Label(type=entity, name=PERSON, span=[0, 3], value=Jim), oneai.Label(type=entity, name=ORG, span=[21, 31], value=Acme Corp.)]
    """

    enrichment: bool = False
    amount: Literal["more", "less", "normal"] = None


@skillclass(api_name="numbers")
class Numbers(Skill):
    """
    Detects and classifies numbers and dates in the input

    ## Output Attributes

    `entities: list[Label]`
        A list of `Label` objects, with detected numbers and their type

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Numbers()
    ... ])
    >>> output = pipeline.run('Jim bought shares of Acme Corp.')
    >>> output.numbers
    """


@skillclass(api_name="sentences")
class SplitBySentence(Skill):
    """
    Splits input by sentence

    ## Output Attributes

    `sentences: list[Label]`
        A list of `Label` objects, with detected numbers and their type

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Sentences()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.sentences
    """


@skillclass(
    api_name="dialogue-segmentation",
    labels_attr="segments",
)
class SplitByTopic(Skill):
    """
    Splits input by discussed topics

    ## Output Attributes

    `segments: list[Label]`
        A list of `Label` objects, with generated topic-segments

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.SplitByTopic()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.segments
    """

    std_ratio: float = None
    amount: Literal["more", "less", "normal"] = None
    use_discourse_parser: bool = False


@skillclass(api_name="sales-insights", labels_attr="sales_insights")
class SalesInsights(Skill):
    """
    Splits input by discussed topics

    ## Output Attributes

    `sales_insights: list[Label]`
        A list of `Label` objects, with detected sales insights

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.SalesInsights()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.sales_insights
    """


@skillclass(
    api_name="service-email-insights",
    labels_attr="service_insights",
)
class ServiceInsights(Skill):
    """
    Finds spans in the text with Service Insights.

    ## Output Attributes

    `service_insights: list[Label]`
        A list of `Label` objects, with detected service insights

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.ServiceInsights()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.service_insights
    """


@skillclass(
    api_name="service-email-insights",
    labels_attr="email_insights",
)
class EmailInsights(Skill):
    """
    Finds spans in the text with Email Insights.

    ## Output Attributes

    `email_insights: list[Label]`
        A list of `Label` objects, with detected email insights

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.EmailInsights()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.email_insights
    """


@skillclass(api_name="detect-language", labels_attr="language")
class DetectLanguage(Skill):
    """
    Detects language of input

    ## Output Attributes

    `language: list[Label]`
        A list of `Label` objects, containing the detected language

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.DetectLanguage()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.language[0]
    """


@skillclass(api_name="headline")
class Headline(Skill):
    """
    Generates a headline based on input

    ## Output Attributes

    `headline: list[Label]`
        A list of `Label` objects, containing the generated headline

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Headline()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.headline[0]
    """


@skillclass(api_name="subheading")
class Subheading(Skill):
    """
    Generates a subheading based on input

    ## Output Attributes

    `subheading: list[Label]`
        A list of `Label` objects, containing the generated subheading

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Subheading()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.subheading[0]
    """


@skillclass(
    api_name="transcribe",
    text_attr="transcription",
    labels_attr="words",
)
class Transcribe(Skill):
    """
    Transcribes audio files

    ## Output

    Speech-to-text transcription of the audio data

    ## Example

    >>> input = oneai.File('./my_audio_file.wav')
    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Transcribe()
    ... ])
    >>> output = pipeline.run(input)
    >>> str(output.transcription.text)[:20] + ' ...'
    oneai.Conversation([ ...
    """

    speaker_detection: bool = True
    timestamp_per_word: bool = False
    engine: Literal["default", "whisper"] = "default"


@skillclass(api_name="clustering")
class Clustering(Skill):
    """
    Automatically organizes various skill labels in clusters for later analytics

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Sentiments(),
    ...     oneai.skills.Clustering(),
    ... ])
    >>> pipeline.run(input)
    """

    collection: str = ""
    """The name of the collection to insert the input into"""
    input_skill: str = ""
    """Use the output of a Skill as input for clustering, omit to use the input directly"""


@skillclass(api_name="gpt", text_attr="comletion")
class GPT(Skill):
    """
    Generate text using GPT-3

    ## Output Attributes

    `completion: str`
        The generated text

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.GPT()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.completion
    """

    model: Literal[
        "text-davinci-003", "text-curie-001", "text-babbage-001", "text-ada-001"
    ] = "text-davinci-003"
    input_skill: str = ""
    prompt: str = ""
    prompt_position: Literal["start", "end"] = "start"
    temperature: float = None
    get_api_key: str = None


@skillclass(api_name="classify", labels_attr="classification")
class Classify(Skill):
    """
    Classifies input based on an analytics collection

    ## Output Attributes

    `classification: list[Label]`
        A list of `Label` objects, containing the classification

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Classify(collection='my_collection')
    ... ])
    >>> output = pipeline.run('...')
    >>> output.classification[0]
    """

    collection: str = ""
    """The name of the collection to use for classification"""
    key: str = ""
    """The key to use for classification"""
    input_skill: str = ""
    """Use the output of a Skill as input for classification, omit to use the input directly"""


class OutputAttrs:
    summary: "Output[str]" = None
    proofread: "Output[List[Utterance]]" = None
    html_article: "Output[str]" = None
    html_text: "Output[str]" = None
    pdf_text: "Output[str]" = None
    transcription: "Output[List[Utterance]]" = None
    anonymized: "Output" = None
    words: Labels = None
    emotions: Labels = None
    names: Labels = None
    numbers: Labels = None
    keywords: Labels = None
    highlights: Labels = None
    sentiments: Labels = None
    topics: Labels = None
    sales_insights: Labels = None
    service_insights: Labels = None
    action_items: Labels = None
    pricing: Labels = None
    replacements: Labels = None
    origins: Labels = None
    segments: Labels = None
    sentences: Labels = None
    anonymizations: Labels = None
    language: Labels = None
    headline: Labels = None
    subheading: Labels = None
    classification: Labels = None



from datetime import timedelta
import json
import urllib.parse
from typing import Awaitable, List

import aiohttp
import oneai, oneai.api
from oneai.api.output import build_output
from oneai.classes import Input, Output, Skill
from oneai.exceptions import handle_unsuccessful_response, validate_api_key

endpoint_default = "api/v0/pipeline"
endpoint_async_file = "api/v0/pipeline/async/file"
endpoint_async_tasks = "api/v0/pipeline/async/tasks"


def build_request(
    input: Input, steps: List[Skill], multilingual: bool, include_text: bool
):
    def json_default(obj):
        if isinstance(obj, timedelta):
            return str(obj)
        if isinstance(obj, Skill):
            return obj.api_name
        return {k: v for k, v in obj.__dict__.items() if v is not None}

    # use input metadata for clustering
    if hasattr(input, "metadata"):
        for skill in steps:
            if skill.api_name == "clustering":
                skill.params["user_metadata"] = input.metadata
                break

    request = {
        "steps": [skill.asdict() for skill in steps],
        "output_type": "json",
        "multilingual": multilingual,
    }
    if include_text:
        request["text"] = input.text
    if isinstance(input, Input):
        if input.type:
            request["input_type"] = input.type
        if hasattr(input, "content_type") and input.content_type:
            request["content_type"] = input.content_type
        if hasattr(input, "encoding") and input.encoding:
            request["encoding"] = input.encoding
    return json.dumps(request, default=json_default)


async def post_pipeline(
    session: aiohttp.ClientSession,
    input: Input,
    steps: List[Skill],
    api_key: str,
    multilingual: bool,
) -> Awaitable[Output]:
    validate_api_key(api_key)

    request = build_request(input, steps, multilingual, True)
    url = f"{oneai.URL}/{endpoint_default}"
    headers = {
        "api-key": api_key,
        "Content-Type": "application/json",
        "User-Agent": f"python-sdk/{oneai.__version__}/{oneai.api.uuid}",
    }

    if oneai.DEBUG_LOG_REQUESTS:
        oneai.logger.debug(f"POST {url}\n")
        oneai.logger.debug(f"headers={json.dumps(headers, indent=4)}\n")
        oneai.logger.debug(f"data={json.dumps(json.loads(request), indent=4)}\n")

    async with session.post(url, headers=headers, data=request) as response:
        if response.status != 200:
            await handle_unsuccessful_response(response)
        else:
            return build_output(steps, await response.json())


async def post_pipeline_async_file(
    session: aiohttp.ClientSession,
    input: Input,
    steps: List[Skill],
    api_key: str,
    multilingual: bool,
) -> Awaitable[str]:
    validate_api_key(api_key)

    request = build_request(input, steps, multilingual, False)
    url = f"{oneai.URL}/{endpoint_async_file}?pipeline=" + urllib.parse.quote(request)
    headers = {
        "api-key": api_key,
        "Content-Type": "application/json",
        "User-Agent": f"python-sdk/{oneai.__version__}/{oneai.api.uuid}",
    }

    if oneai.DEBUG_LOG_REQUESTS:
        oneai.logger.debug(f"POST {url}\n")
        oneai.logger.debug(f"headers={json.dumps(headers, indent=4)}\n")
        oneai.logger.debug(f"data={json.dumps(json.loads(request), indent=4)}\n")

    async with session.post(url, headers=headers, data=input.text) as response:
        if response.status != 200:
            await handle_unsuccessful_response(response)
        else:
            return await response.json()


async def get_task_status(
    session: aiohttp.ClientSession,
    task_id: str,
    api_key: str,
):
    validate_api_key(api_key)

    url = f"{oneai.URL}/{endpoint_async_tasks}/{task_id}"
    headers = {
        "api-key": api_key,
        "User-Agent": f"python-sdk/{oneai.__version__}/{oneai.api.uuid}",
    }

    if oneai.DEBUG_LOG_REQUESTS:
        oneai.logger.debug(f"GET {url}\n")
        oneai.logger.debug(f"headers={json.dumps(headers, indent=4)}\n")

    async with session.get(url, headers=headers) as response:
        if response.status != 200:
            await handle_unsuccessful_response(response)
        else:
            return await response.json()



import asyncio
from datetime import datetime, timedelta
import logging
from typing import Awaitable, Callable, Iterable, List

import aiohttp

import oneai
from oneai.api.output import build_output
from oneai.api.pipeline import post_pipeline, post_pipeline_async_file, get_task_status
from oneai.classes import Input, Output, PipelineInput, Skill
from oneai.exceptions import ServerError, handle_unsuccessful_response

logger = logging.getLogger("oneai")


def time_format(time: timedelta):
    minutes = f"{time.seconds // 60}m " if time.seconds > 59 else ""
    return minutes + f"{time.seconds % 60}s {time.microseconds // 1000}ms"


STATUS_COMPLETED = "COMPLETED"
STATUS_FAILED = "FAILED"


# open a client session and send a request
async def process_single_input(
    input: PipelineInput, steps: List[Skill], api_key: str, multilingual: bool = False
) -> Awaitable[Output]:
    timeout = aiohttp.ClientTimeout(total=6000)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        return await _run_internal(
            session, Input.wrap(input), steps, api_key, multilingual
        )


async def process_file_async(
    input: PipelineInput,
    steps: List[Skill],
    api_key: str,
    interval: int,
    multilingual: bool = False,
) -> Awaitable[Output]:
    input = Input.wrap(input, False)
    timeout = aiohttp.ClientTimeout(total=6000)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        name = input.text.name
        logger.debug(f"Uploading file '{name}'")
        task_id = (
            await post_pipeline_async_file(session, input, steps, api_key, multilingual)
        )["task_id"]
        logger.debug(f"Upload of file '{name}' complete\n")

        status = ""
        start = datetime.now()
        while status != STATUS_COMPLETED:
            if status == STATUS_FAILED:
                await handle_unsuccessful_response(response["result"])
            response = await get_task_status(session, task_id, api_key)
            status = response["status"]
            logger.debug(
                f"Processing file '{name}' - status {status} - {time_format(datetime.now() - start)}"
            )
            await asyncio.sleep(interval)
        logger.debug(
            f"Processing of file '{name}' complete - {time_format(datetime.now() - start)} total\n"
        )
        return build_output(steps, response["result"])


# open a client session with multiple workers and send concurrent requests
async def process_batch(
    batch: Iterable[PipelineInput],
    steps: List[Skill],
    on_output: Callable[[PipelineInput, Output], None],
    on_error: Callable[[PipelineInput, Exception], None],
    api_key: str,
    multilingual: bool = False,
):
    iterator = iter(batch)
    successful = 0  # total successful responses
    failed = 0  # number of exceptions occurred
    time_total = timedelta()  # total time spent on all requests
    # length = len(batch) if hasattr(batch, "__len__") else 0

    def next_input():  # distribute batch to workers
        try:
            return Input.wrap(next(iterator))
        except StopIteration:
            return None  # we need to break loop for each worker, so we ignore StopIteration

    def log_progress(
        time_delta=timedelta(), start=False, end=False
    ):  # todo progress bar for iterables with __len__
        nonlocal successful, failed, time_total

        time_total += time_delta
        if start:
            logger.debug(
                f"Starting batch processing with {oneai.MAX_CONCURRENT_REQUESTS} workers"
            )
        elif end:
            logger.debug(
                "Processed %d inputs - %s/input - %s total - %d successful - %d failed\n"
                % (
                    successful + failed,
                    time_format(
                        time_total
                        / (successful + failed)
                        / oneai.MAX_CONCURRENT_REQUESTS
                    ),
                    time_format(time_total / oneai.MAX_CONCURRENT_REQUESTS),
                    successful,
                    failed,
                )
            )
        else:
            logger.debug(
                "Input %d - %s/input - %s total - %d successful - %d failed"
                % (
                    successful + failed,
                    time_format(time_delta),
                    time_format(time_total / oneai.MAX_CONCURRENT_REQUESTS),
                    successful,
                    failed,
                )
            )

    async def req_worker(session):  # run requests sequentially
        nonlocal successful, failed

        time_start = datetime.now()
        input = next_input()
        while input:
            try:
                output = await _run_internal(
                    session, input, steps, api_key, multilingual
                )
                on_output(input, output)
                successful += 1
            except Exception as e:  # todo: break loop for some error types
                logger.error(f"Input {successful + failed}: {repr(e)}")
                on_error(input, e)
                failed += 1

            time_end = datetime.now()
            log_progress(time_end - time_start)
            time_start = time_end
            input = next_input()

    workers = []
    timeout = aiohttp.ClientTimeout(total=6000)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        for _ in range(oneai.MAX_CONCURRENT_REQUESTS):
            worker = asyncio.create_task(req_worker(session))
            workers.append(worker)
        log_progress(start=True)
        await asyncio.gather(*workers)
        log_progress(end=True)


async def fetch_url(session: aiohttp.ClientSession, url: str):
    async with session.get(url) as response:
        if response.status != 200:
            raise ServerError(
                50001,
                "Retrieve URL failed",
                f"Failed to retrieve the input from url '{url}'.",
            )
        return Input(await response.text(), type="article", content_type="text/plain")


async def _run_internal(
    session: aiohttp.ClientSession,
    input: Input,
    skills: List[Skill],
    api_key: str,
    multilingual: bool,
) -> Awaitable[Output]:
    if not skills:  # no skills
        return Output(input.text)

    if input.content_type == "text/uri-list":
        input = await fetch_url(session, input.text)

    return await post_pipeline(session, input, skills, api_key, multilingual)
